#define PILOT_PRECOMPILED_HEADERS_OFF
#include <PalmOS.h>
#include "3dpongRsc.h"
#include <SysEvtMgr.h>
#include <PenMgr.h>
#include <KeyMgr.h>

Err RomVersionCompatible(int requiredVersion, int launchFlags);
void* GetObjectPtr(int objectID);
Boolean MainFormHandleEvent(EventPtr eventP);
Boolean AppHandleEvent( EventPtr eventP);

typedef struct 
{
	char replaceme;
} PongPreferenceType;

typedef struct 
{
	char replaceme;
} PongAppInfoType;

typedef PongAppInfoType* PongAppInfoPtr;

typedef struct
{
	int X, Y, Z;
} Vector3DType;

Vector3DType ZeroVector3D = { 0, 0, 0 };
	
typedef struct
{
	int Top, Left, Right, Bottom;
} RectType;
	
typedef struct
{
	Vector3DType Pos;
	Vector3DType Vel;
//	Vector3d Spin; TODO
} BallType;
	
typedef struct
{
	// Center of the paddle
	Vector3DType Pos;
	
	// Physics stuff
	Vector3DType Delta;   // Amount moved since last update for spin calc.
	int HalfWidth;
	int HalfHeight;

	// Stuff for moving the paddle forward.
	int TargetZ;						
	int DefaultZ;						
	int ForwardZ;						

	// AI stuff
	Vector3DType Vel;
	int Speed;

	// Game stuff
	int Score;
} PaddleType;

typedef struct
{
	// How long the stage is from near side to far side.
	int Depth;
	RectType Window;	

	int Gravity;
	int BallSpeed;
} StageType;

typedef struct 
{
	int Speed;
	Boolean Recenter;
} AIType;

typedef struct
{
	int NextPeriodTime;
	int EndTimeout;

	AIType AI;

	StageType Stage;
	BallType Ball;
	PaddleType Paddle1, Paddle2;
	
	// Single frame trailing effect for better visibility.
	BallType OldBall;
	PaddleType OldPaddle1, OldPaddle2; 

	// Score variable from last frame, to see if we need to erase the scoring graphic.
	int LastScore;
	
	WinHandle OffscreenWindow;
} GameStatusType;

GameStatusType Game;

// Direct screen access
Coord DisplayWidth = 160;
Coord DisplayHeight = 160;
int DisplayCenterX = 80;
int DisplayCenterY = 80;

#define appFileCreator            'wadb'
#define appVersionNum             0x01
#define appPrefID                 0x00
#define appPrefVersionNum         0x01

// Define the minimum OS version we support
#define ourMinVersion sysMakeROMVersion(2,0,0,sysROMStageRelease,0)

// Game constants
#define GamePeriod    2
#define ViewportScale 100

#define ScreenWidth   160
#define ScreenHeight  145
#define ScreenTop     13
#define	ScreenLeft    0

#define TimeRes       32 // Time resolution for ball physics.

#define BallSize      5 // Size of the ball (actually radius)

int clamp(int a, int b, int c)
{
	if (a<b) return b;
	else if (a>c) return c;
	else return a;
}

int GetProjectedX( int X, int Y, int Z )
{
	return DisplayCenterX + ( X - DisplayCenterX ) * ViewportScale / ( Z + ViewportScale );
}

int GetProjectedY( int X, int Y, int Z )
{
	return DisplayCenterY + ( Y - DisplayCenterY ) * ViewportScale / ( Z + ViewportScale );
}

void DrawLine3D( int X1, int Y1, int Z1, int X2, int Y2, int Z2, Boolean draw )
{
	if ( draw )
		WinDrawLine(
			GetProjectedX( X1, Y1, Z1 ), GetProjectedY( X1, Y1, Z1 ),
			GetProjectedX( X2, Y2, Z2 ), GetProjectedY( X2, Y2, Z2 ) );
	else
		WinEraseLine(
			GetProjectedX( X1, Y1, Z1 ), GetProjectedY( X1, Y1, Z1 ),
			GetProjectedX( X2, Y2, Z2 ), GetProjectedY( X2, Y2, Z2 ) );
}

void DrawRect3D( RectType* Rect, int Depth, Boolean draw )
{
	RectangleType Temp;
	
	Temp.topLeft.x = GetProjectedX( Rect->Left, Rect->Top, Depth ) + 1;
	Temp.topLeft.y = GetProjectedY( Rect->Left, Rect->Top, Depth ) + 1;
	Temp.extent.x = GetProjectedX( Rect->Right, Rect->Bottom, Depth ) - Temp.topLeft.x - 1;
	Temp.extent.y = GetProjectedY( Rect->Right, Rect->Bottom, Depth ) - Temp.topLeft.y - 1;

	if ( draw )
		WinDrawRectangleFrame( rectangleFrame, &Temp );
	else
		WinEraseRectangleFrame( rectangleFrame, &Temp );
		
}

void DrawFilledRect3D( RectType* Rect, int Depth, Boolean draw )
{
	RectangleType Temp;
	
	Temp.topLeft.x = GetProjectedX( Rect->Top, Rect->Left, Depth ) + 1;
	Temp.topLeft.y = GetProjectedY( Rect->Top, Rect->Left, Depth ) + 1;
	Temp.extent.x = GetProjectedX( Rect->Right, Rect->Bottom, Depth ) - Temp.topLeft.x - 1;
	Temp.extent.y = GetProjectedY( Rect->Right, Rect->Bottom, Depth ) - Temp.topLeft.y - 1;

	if ( draw )
		WinDrawRectangle( &Temp, 0 );
	else
		WinEraseRectangle( &Temp, 0 );
		
}

void DrawStage( StageType* Stage, Boolean draw )
{
	RectType* Window = &Stage->Window;

	// Near and far rectangles	
	DrawRect3D( Window, 0, draw );
	DrawRect3D( Window, Stage->Depth, draw );

	// Diagonals
	DrawLine3D( Window->Left, Window->Top, 1, Window->Left, Window->Top, Stage->Depth, draw );
	DrawLine3D( Window->Left, Window->Bottom, 1, Window->Left, Window->Bottom, Stage->Depth, draw );
	DrawLine3D( Window->Right, Window->Top, 1, Window->Right, Window->Top, Stage->Depth, draw );
	DrawLine3D( Window->Right, Window->Bottom, 1, Window->Right, Window->Bottom, Stage->Depth, draw );
}

void DrawPaddle( PaddleType* Paddle, StageType* Stage, Boolean draw )
{
	RectType Temp;
	Temp.Left = Paddle->Pos.X - Paddle->HalfWidth;	
	Temp.Right = Paddle->Pos.X + Paddle->HalfWidth;	
	Temp.Top = Paddle->Pos.Y - Paddle->HalfHeight;	
	Temp.Bottom = Paddle->Pos.Y + Paddle->HalfHeight;	
	
	DrawRect3D( &Temp, Paddle->Pos.Z, draw );
	
	DrawLine3D( 
		Temp.Left + ( ( Temp.Right - Temp.Left ) >> 1 ),
		Temp.Bottom,
		Paddle->Pos.Z,
		Temp.Left + ( ( Temp.Right - Temp.Left ) >> 1 ),
		Stage->Window.Bottom,
		Paddle->Pos.Z, draw );
}

void DrawBall( BallType* Ball, StageType* Stage, Boolean draw )
{
	RectType Temp;
	Temp.Left = Ball->Pos.X - BallSize;	
	Temp.Right = Ball->Pos.X + BallSize;	
	Temp.Top = Ball->Pos.Y - BallSize;	
	Temp.Bottom = Ball->Pos.Y + BallSize;	
	
	// Draw the ball.
	DrawRect3D( &Temp, Ball->Pos.Z, draw );
	
	// Draw the shadow.
	DrawLine3D( Temp.Left, Stage->Window.Bottom, Ball->Pos.Z, Temp.Right, Stage->Window.Bottom, Ball->Pos.Z, draw );
}

// Check paddle inputs and apply to Paddle.
void UpdatePaddleFromPen ( PaddleType* Paddle, StageType* Stage )
{
	Int16 PenX, PenY;
	Boolean PenDown;
	int KeyState;
	
	Vector3DType LastPos;
	
	LastPos = Paddle->Pos;
	
	EvtGetPen( &PenX, &PenY, &PenDown );

	// Get the current state of the hardware keys.
	KeyState = KeyCurrentState();

	if ( KeyState & keyBitPageUp )
		Paddle->TargetZ = Paddle->ForwardZ;
	else
		Paddle->TargetZ = Paddle->DefaultZ;

	// Snaps forward, eases back.
	if ( Paddle->Pos.Z < Paddle->TargetZ )
	{
		if ( Paddle->Delta.Z < 4 ) 
			Paddle->Delta.Z = 6;
		Paddle->Pos.Z += Paddle->Delta.Z + 2;
		if ( Paddle->Pos.Z > Paddle->TargetZ )
			Paddle->Pos.Z = Paddle->TargetZ;
	}

	if ( Paddle->Pos.Z > Paddle->TargetZ )
		Paddle->Pos.Z += ( Paddle->TargetZ - Paddle->Pos.Z ) / 4;
		
	// Get the 2d position from the pen.
	if ( PenDown ) 
	{
		if ( PenY > 160 )
		{
			// Its in the graffiti area, so interpret it in the correct way (scaled to the visible area).
			Paddle->Pos.X = ( PenX - 20 ) * 160 / 120;
			Paddle->Pos.Y = ( PenY - 170 ) * 270 / 100;
		}
		else
		{
			Paddle->Pos.X = PenX;
			Paddle->Pos.Y = PenY - ScreenTop;
		}
		
		// Clip the paddle position.
		if ( Paddle->Pos.X < Paddle->HalfWidth )
			Paddle->Pos.X = Paddle->HalfWidth;
			
		if ( Paddle->Pos.Y < Paddle->HalfHeight ) 
			Paddle->Pos.Y = Paddle->HalfHeight;
			
		if ( Paddle->Pos.X > Stage->Window.Right - Paddle->HalfWidth )
			Paddle->Pos.X = Stage->Window.Right - Paddle->HalfWidth;
			
		if ( Paddle->Pos.Y > Stage->Window.Bottom - Paddle->HalfHeight )
			Paddle->Pos.Y = Stage->Window.Bottom - Paddle->HalfHeight;
	}
	
	Paddle->Delta.X = Paddle->Pos.X - LastPos.X;
	Paddle->Delta.Y = Paddle->Pos.Y - LastPos.Y;
	Paddle->Delta.Z = Paddle->Pos.Z - LastPos.Z;
}

// Compute AI and move Paddle.
void UpdatePaddleAI( PaddleType* Paddle, BallType* Ball, StageType* Stage )
{
	// Only move when the ball is coming back, that way it appears to react to the players hit.
	// Actually, start moving just before the player hits it.
	if ( Ball->Vel.Z > 0 || Ball->Vel.Z < 0 && Ball->Pos.Z < 30) 
	{
		// Acceleration towards the ball.
		if ( Abs( ( Paddle->Pos.X - Ball->Pos.X ) ) > 5 )
		{
			if ( Paddle->Pos.X < Ball->Pos.X )
				Paddle->Vel.X+=2;
			if ( Paddle->Pos.X > Ball->Pos.X ) 
				Paddle->Vel.X-=2;
		}
			
		if ( Abs( ( Paddle->Pos.Y - Ball->Pos.Y ) ) > 5 )
		{
			if ( Paddle->Pos.Y < Ball->Pos.Y )
				Paddle->Vel.Y+=2;
			if ( Paddle->Pos.Y > Ball->Pos.Y ) 
				Paddle->Vel.Y-=2;
		}
		
		// Speed clamping
		Paddle->Vel.X = clamp( Paddle->Vel.X, -Game.AI.Speed, Game.AI.Speed );
		Paddle->Vel.Y = clamp( Paddle->Vel.Y, -Game.AI.Speed, Game.AI.Speed );
	}
	else
	if ( Ball->Pos.Z < 110 )
	{
		Paddle->Vel.X = 0;
		Paddle->Vel.Y = 0;
		// Drift towards the center.
		if ( Game.AI.Recenter )
		{
			Paddle->Pos.X += ( 80 - Paddle->Pos.X ) / 4;
			Paddle->Pos.Y += ( 80 - Paddle->Pos.Y ) / 4;
		}
	}
			
	// Friction
	if ( Paddle->Vel.X > 0 )
		Paddle->Vel.X--;
	if ( Paddle->Vel.X < 0 )
		Paddle->Vel.X++;
		
	if ( Paddle->Vel.Y > 0 )
		Paddle->Vel.Y--;
	if ( Paddle->Vel.Y < 0 )
		Paddle->Vel.Y++;
		
	Paddle->Pos.X += Paddle->Vel.X;
	Paddle->Pos.Y += Paddle->Vel.Y;
	
	// Clip the paddle position
	if ( Paddle->Pos.X < Paddle->HalfWidth )
		Paddle->Pos.X = Paddle->HalfWidth;
		
	if ( Paddle->Pos.Y < Paddle->HalfHeight ) 
		Paddle->Pos.Y = Paddle->HalfHeight;
		
	if ( Paddle->Pos.X > Stage->Window.Right - Paddle->HalfWidth )
		Paddle->Pos.X = Stage->Window.Right - Paddle->HalfWidth;
		
	if ( Paddle->Pos.Y > Stage->Window.Bottom - Paddle->HalfHeight )
		Paddle->Pos.Y = Stage->Window.Bottom - Paddle->HalfHeight;
}

// 0 if nobody scored, 1 if Paddle1 scored, 2 if Paddle2 scored.
int UpdateBall ( BallType* Ball, PaddleType* Paddle1, PaddleType* Paddle2, StageType* Stage )
{
	// Ball collisions are handled very accurately, as this is the basis of the game.
	// All times are in 1sec/TimeRes units.
	// 1. Loop through all the surfaces and finds the first one the ball will collide with
	// in this animation frame (if any). 
	// 2. Update the Ball velocity based on the collision, and advance the current time to 
	// the exact time of the collision.
	// 3. Goto step 1, until no collisions remain in the current animation frame.

	int TimeLeft = TimeRes;					// Time remaining in this animation frame.
	int FirstCollisionTime;					// -1 means no collision found.
	Vector3DType FirstCollisionVel;	// New ball velocity from first collision.
	int FirstCollisionType;					// 0 for normal collision (wall), otherwise the scorezone number hit. 
	
	Vector3DType NextBallPos;				// Hypothetical ball position assuming no collision.
	int	Time;												// Time of current collision.
	
	int CollisionType = 0;					// Stored return value.
	int HitWall = 0;
	
	int Iterations = 0;

	do {
		Iterations++;
		if ( Iterations > 5 )
			break;

		// Calculate new next ball position.
		NextBallPos = Ball->Pos;
		NextBallPos.X += Ball->Vel.X * TimeLeft / TimeRes;
		NextBallPos.Y += Ball->Vel.Y * TimeLeft / TimeRes;
		NextBallPos.Z += Ball->Vel.Z * TimeLeft / TimeRes;

		// Reset FirstCollisionTime.		
		FirstCollisionTime = -1;
		
		// Check stage walls.  First checks to see if the boundary was crossed, if so then calculates time, etc.
		if ( NextBallPos.X - BallSize <= 0 ) // Left wall
		{
			Time = ( Ball->Pos.X - BallSize ) * TimeRes / -Ball->Vel.X;	// negative VX is to account for left wall facing.
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = -Ball->Vel.X;
				FirstCollisionVel.Y = Ball->Vel.Y;
				FirstCollisionVel.Z = Ball->Vel.Z;
				FirstCollisionType = 0;
			}
		}
		if ( NextBallPos.X + BallSize >= Stage->Window.Right ) // Right wall
		{
			Time = ( Stage->Window.Right - ( Ball->Pos.X + BallSize ) ) * TimeRes / Ball->Vel.X;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = -Ball->Vel.X;
				FirstCollisionVel.Y = Ball->Vel.Y;
				FirstCollisionVel.Z = Ball->Vel.Z;
				FirstCollisionType = 0;
			}
		}
		if ( NextBallPos.Y - BallSize <= 0 ) // Top wall
		{
			Time = ( Ball->Pos.Y - BallSize ) * TimeRes / -Ball->Vel.Y;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = Ball->Vel.X;
				FirstCollisionVel.Y = -Ball->Vel.Y;
				FirstCollisionVel.Z = Ball->Vel.Z;
				FirstCollisionType = 0;
			}
		}
		if ( NextBallPos.Y + BallSize >= Stage->Window.Bottom ) // Bottom wall
		{
			Time = ( Stage->Window.Bottom - ( Ball->Pos.Y + BallSize ) ) * TimeRes / Ball->Vel.Y;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = Ball->Vel.X;
				FirstCollisionVel.Y = -Ball->Vel.Y;
				FirstCollisionVel.Z = Ball->Vel.Z;
				FirstCollisionType = 0;
			}
		}
		if ( NextBallPos.Z <= 0 ) // Front wall
		{
			Time = Ball->Pos.Z * TimeRes / -Ball->Vel.Z;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = (SysRandom(0)%3)-1 * Game.Stage.BallSpeed;
				FirstCollisionVel.Y = (SysRandom(0)%3)-1 * Game.Stage.BallSpeed;
				FirstCollisionVel.Z = Game.Stage.BallSpeed;
				FirstCollisionType = 2;
			}
		}
		if ( NextBallPos.Z >= Stage->Depth ) // Back wall
		{
			Time = ( Stage->Depth - Ball->Pos.Z ) * TimeRes / Ball->Vel.Z;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = (SysRandom(0)%3)-1 * Game.Stage.BallSpeed;
				FirstCollisionVel.Y = (SysRandom(0)%3)-1 * Game.Stage.BallSpeed;
				FirstCollisionVel.Z = -Game.Stage.BallSpeed;
				FirstCollisionType = 1;
			}
		}
		// Paddle collision.  Paddle collisions are inaccurate, in that it doesn't take into account the velocity of 
		// the ball in its 2D check, it uses the original 2D position.
		if ( 		Ball->Vel.Z < 0 
				&&	( Ball->Pos.Z >= Paddle1->Pos.Z || Ball->Pos.Z >= Paddle1->Pos.Z - Paddle1->Delta.Z ) 
				&& 	( NextBallPos.Z <= Paddle1->Pos.Z || NextBallPos.Z <= Paddle1->Pos.Z - Paddle1->Delta.Z )
				&&	Ball->Pos.X >= Paddle1->Pos.X - Paddle1->HalfWidth
				&&	Ball->Pos.X <= Paddle1->Pos.X + Paddle1->HalfWidth 
				&&	Ball->Pos.Y >= Paddle1->Pos.Y - Paddle1->HalfHeight
				&&	Ball->Pos.Y <= Paddle1->Pos.Y + Paddle1->HalfHeight )
		{
			Time = ( Ball->Pos.Z - Paddle1->Pos.Z ) * TimeRes / -Ball->Vel.Z;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = Ball->Vel.X;
				FirstCollisionVel.Y = Ball->Vel.Y;
				FirstCollisionVel.Z = -Ball->Vel.Z;

				// If paddle is moving forward, bounce the ball off.
				if ( Paddle1->Delta.Z > 0 )
				{
					FirstCollisionVel.Z += 4;

					// Apply some pong like angling based on where it hits the paddle.
					if ( NextBallPos.X - Paddle1->Pos.X > 20 )
  						FirstCollisionVel.X += 2;
					if ( NextBallPos.X - Paddle1->Pos.X < -20 )
						FirstCollisionVel.X -= 2;

					if ( NextBallPos.Y - Paddle1->Pos.Y > 15 )
						FirstCollisionVel.Y += 2;
					if ( NextBallPos.Y - Paddle1->Pos.Y < -15 )
						FirstCollisionVel.Y -= 2;
				}
				// Likewise if paddle is moving backwards, cushion it.
				if ( Paddle1->Delta.Z < 0 )
					FirstCollisionVel.Z -= 2;
				
				FirstCollisionType = 0;
			}
		}
		// Computer paddle.
		if ( 		Ball->Vel.Z > 0 
				&&	( Ball->Pos.Z <= Paddle2->Pos.Z || Ball->Pos.Z <= Paddle2->Pos.Z - Paddle2->Delta.Z ) 
				&& 	( NextBallPos.Z >= Paddle2->Pos.Z || NextBallPos.Z >= Paddle2->Pos.Z - Paddle2->Delta.Z )
				&&	Ball->Pos.X >= Paddle2->Pos.X - Paddle2->HalfWidth
				&&	Ball->Pos.X <= Paddle2->Pos.X + Paddle2->HalfWidth 
				&&	Ball->Pos.Y >= Paddle2->Pos.Y - Paddle2->HalfHeight
				&&	Ball->Pos.Y <= Paddle2->Pos.Y + Paddle2->HalfHeight ) // Paddle 2
		{
			Time = ( Paddle2->Pos.Z - Ball->Pos.Z ) * TimeRes / Ball->Vel.Z;
			if ( FirstCollisionTime == -1 || Time < FirstCollisionTime )
			{
				// Set new first collision.
				FirstCollisionTime = Time;
				FirstCollisionVel.X = Ball->Vel.X;
				FirstCollisionVel.Y = Ball->Vel.Y;
				FirstCollisionVel.Z = -Ball->Vel.Z + ( Paddle1->Delta.Z > 0 ) * 2 + ( Paddle1->Delta.Z < 0 ) * 2;
				FirstCollisionType = 0;
			}
		}

		// Advance the ball to the point of the first collision.
		if ( FirstCollisionTime != -1 )		
		{
			if ( FirstCollisionType == 0 )
				HitWall = 1;
			
			Ball->Pos.X += Ball->Vel.X * FirstCollisionTime / TimeRes;
			Ball->Pos.Y += Ball->Vel.Y * FirstCollisionTime / TimeRes;
			Ball->Pos.Z += Ball->Vel.Z * FirstCollisionTime / TimeRes;
			Ball->Vel = FirstCollisionVel;
			
			TimeLeft -= FirstCollisionTime;
			CollisionType = FirstCollisionType;
		}
	} while ( FirstCollisionTime != -1 && TimeLeft > 0 );

	// If there's time left in the frame w/o collision, finish it up.	
	if ( TimeLeft > 0 )
	{
		Ball->Pos.X += Ball->Vel.X * TimeLeft / TimeRes;
		Ball->Pos.Y += Ball->Vel.Y * TimeLeft / TimeRes;
		Ball->Pos.Z += Ball->Vel.Z * TimeLeft / TimeRes;
	}

	// Apply gravity.
	Ball->Vel.Y += Game.Stage.Gravity;
	if ( Ball->Pos.Y + BallSize + 20 > Stage->Window.Bottom && abs(Ball->Vel.Y) == 0 )
		Ball->Vel.Y -= 6;

	// Calculate scores if any collisions with the back wall happened.
	if ( CollisionType == 1 )
	{
		Paddle1->Score += 10;
		SndPlaySystemSound( sndError );
	}
	else
	if ( CollisionType == 2 )
	{
		Paddle2->Score += 10;	
		SndPlaySystemSound( sndError );
	}
	else
		SndPlaySystemSound( sndClick );
		
	return CollisionType;
}

void GameStateInit (void)
{
	RectangleType bounds;

	// Setup the screenBuffer.  Clear it for use.
	WinSetDrawWindow(Game.OffscreenWindow);
	bounds.topLeft.x = 0;
	bounds.topLeft.y = 0;
	bounds.extent.x = ScreenWidth;
	bounds.extent.y = ScreenHeight;
	WinEraseRectangle(&bounds, 0);

	Game.NextPeriodTime = TimGetTicks() + GamePeriod;
}

void InitStage( int depth, int grav, int speed )
{
	Game.Stage.Gravity = grav;
	Game.Stage.BallSpeed = speed;
	Game.Stage.Depth = depth;
	Game.Stage.Window.Left = 0;
	Game.Stage.Window.Right = ScreenWidth - 1;
	Game.Stage.Window.Top = 0;
	Game.Stage.Window.Bottom = ScreenHeight - 1;

	Game.Ball.Pos.X = 80;
	Game.Ball.Pos.Y = 80;
	Game.Ball.Pos.Z = 80;
	Game.Ball.Vel.X = (SysRandom(0)%3)-1 * speed;
	Game.Ball.Vel.Y = (SysRandom(0)%3)-1 * speed;
	Game.Ball.Vel.Z = speed;
}

void InitPlayerPaddle( int w, int h )
{
	Game.Paddle1.Score = 0;
	Game.Paddle1.HalfWidth = w;
	Game.Paddle1.HalfHeight = h;
	Game.Paddle1.Pos.X = 80;
	Game.Paddle1.Pos.Y = 80;
	Game.Paddle1.Pos.Z = 10;
	Game.Paddle1.DefaultZ = Game.Paddle1.Pos.Z;
	Game.Paddle1.TargetZ = Game.Paddle1.Pos.Z;
	Game.Paddle1.ForwardZ = 40;
	Game.Paddle1.Delta = ZeroVector3D;

	Game.OldPaddle1 = Game.Paddle1;
}

void InitAIPaddle( int w, int h )
{
	Game.Paddle2.Score = 0;
	Game.Paddle2.HalfWidth = w;
	Game.Paddle2.HalfHeight = h;
	Game.Paddle2.Pos.X = 80;
	Game.Paddle2.Pos.Y = 80;
	Game.Paddle2.Pos.Z = Game.Stage.Depth - 10;
	Game.Paddle2.DefaultZ = Game.Paddle2.Pos.Z;
	Game.Paddle2.TargetZ = Game.Paddle2.Pos.Z;
	Game.Paddle2.ForwardZ = Game.Stage.Depth - 40;
	Game.Paddle2.Delta = ZeroVector3D;
	Game.Paddle2.Vel = ZeroVector3D;

	Game.OldPaddle2 = Game.Paddle2;
}

void CopyBuffers()
{
	RectangleType Bounds;
	// Copy the offscreen buffer to the display.
	Bounds.topLeft.x = 0;
	Bounds.topLeft.y = 0;
	Bounds.extent.x = ScreenWidth;
	Bounds.extent.y = ScreenHeight;
	WinCopyRectangle( Game.OffscreenWindow, WinGetDisplayWindow(), &Bounds, ScreenLeft, ScreenTop, winPaint );
}

int CurLevel = 0;

void NextLevel()
{
	char work[32];
	CurLevel++;
	CopyBuffers();
	StrIToA( work, CurLevel );
	FrmCustomAlert( LevelAlert, work, work, work );
	GameStateInit();
	switch ( CurLevel )
	{
	case 1:
		Game.AI.Speed = 3;
		Game.AI.Recenter = 0;
		InitStage( 160, 0, 3 );
		InitPlayerPaddle( 40, 30 );
		InitAIPaddle( 40, 30 );
		break;
	case 2:
		Game.AI.Speed = 5;
		Game.AI.Recenter = 1;
		InitStage( 160, 0, 4 );
		InitPlayerPaddle( 30, 20 );
		InitAIPaddle( 30, 20 );
		break;
	case 3:
		Game.AI.Speed = 10;
		Game.AI.Recenter = 1;
		InitStage( 160, 0, 5 );
		InitPlayerPaddle( 15, 10 );
		InitAIPaddle( 15, 10 );
		break;
	case 4:
		Game.AI.Speed = 8;
		Game.AI.Recenter = 1;
		InitStage( 160, 1, 5 );
		InitPlayerPaddle( ScreenWidth/2, 15 );
		InitAIPaddle( ScreenWidth/2, 15 );
		break;
	case 5:
		Game.AI.Speed = 8;
		Game.AI.Recenter = 1;
		InitStage( 80, 0, 4 );
		InitPlayerPaddle( 40, 30 );
		InitAIPaddle( 40, 30 );
		break;
	default:
		Game.AI.Speed = CurLevel+8;
		Game.AI.Recenter = 10;
		InitStage( 160, 1, CurLevel+2 );
		InitPlayerPaddle( 40, 30 );
		InitAIPaddle( 40, 30 );
		break;
	}
}

void NewGame()
{
	CurLevel = 0;
	NextLevel();
}

void GameOver()
{
	FrmAlert( GameOverAlert );
	NewGame();
}

void GameStateAdvance (void)
{
//	char work[40];
	WinHandle OldWindow;
	int Score;

	int KeyState = KeyCurrentState();
	if ( KeyState & keyBitPageDown )
		Game.Paddle1.Score = 50;

	// Update and draw the user controlled paddle (for minimum input-visual latency).
	OldWindow = WinSetDrawWindow( Game.OffscreenWindow );
	Game.NextPeriodTime = TimGetTicks() + GamePeriod;
	UpdatePaddleFromPen( &Game.Paddle1, &Game.Stage );
	DrawPaddle( &Game.Paddle1, &Game.Stage, 1 );
	WinSetDrawWindow( OldWindow );
	
	CopyBuffers();

	// Erase moving things from the offscreen buffer.
	// Things get a one-frame trailing effect, to make them more visible.
	OldWindow = WinSetDrawWindow( Game.OffscreenWindow );
	DrawPaddle(&Game.OldPaddle1, &Game.Stage, false);
	Game.OldPaddle1 = Game.Paddle1;
	DrawPaddle(&Game.OldPaddle2, &Game.Stage, false);
	Game.OldPaddle2 = Game.Paddle2;
	DrawBall(&Game.OldBall, &Game.Stage, false);
	Game.OldBall = Game.Ball;

	// If someone scored last frame, wipe the whole screen and start fresh.
	if ( Game.LastScore != 0 )
		DrawFilledRect3D( &Game.Stage.Window, 0, false );	
	
	// Update the game.
	UpdatePaddleAI( &Game.Paddle2, &Game.Ball, &Game.Stage );
	
	Score = UpdateBall( &Game.Ball, &Game.Paddle1, &Game.Paddle2, &Game.Stage );
	if ( Score == 0 )
	{
		// Draw the non-latency critical stuff for the next frame to the offscreen buffer.
		DrawStage( &Game.Stage, 1 );
		DrawBall( &Game.Ball, &Game.Stage, 1 );
	
		DrawPaddle( &Game.Paddle2, &Game.Stage, 1 );
	}
	else
	if ( Score == 1 )
		DrawFilledRect3D( &Game.Stage.Window, Game.Stage.Depth, 1 );	
	if ( Score == 2 )
		DrawFilledRect3D( &Game.Stage.Window, 0, 1 );	

	Game.LastScore = Score;
	
	// Draw the score
	{
		char work[32];
		StrIToA(work, Game.Paddle1.Score );
		WinDrawChars( work, StrLen(work), ScreenWidth/2 - 20, 3 );
		StrIToA(work, Game.Paddle2.Score );
		WinDrawChars( work, StrLen(work), ScreenWidth/2 + 10, 3 );
		StrPrintF(work, "Stage %d", CurLevel );
		WinDrawChars( work, StrLen(work), 5, 3 );
	}
	
	WinSetDrawWindow( OldWindow );

	if ( Game.Paddle1.Score == 50 || Game.Paddle2.Score == 50 )
		Game.EndTimeout++;
	else
		Game.EndTimeout = 0;
	if ( Game.EndTimeout == 3 )
	{
		if ( Game.Paddle2.Score == 50 )
			GameOver();
		if ( Game.Paddle1.Score == 50 )
			NextLevel();
	}
}

void MainFormInit(FormPtr frmP)
{
	WinGetDisplayExtent( &DisplayWidth, &DisplayHeight );
}

Boolean MainFormHandleEvent(EventType *event)
{
	Boolean   handled = false;
	FormType  *form;

	switch (event->eType)
	{
	case frmOpenEvent:
		form = FrmGetActiveForm();
		FrmDrawForm(form);
		handled = 1;
		break;

	default:
		break;
	}

	return handled;
}


Boolean AppHandleEvent( EventPtr eventP)
{
	int formId;
	FormPtr frmP;

	if (eventP->eType == frmLoadEvent)
	{
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		switch (formId)
		{
		case MainForm:
			FrmSetEventHandler(frmP, MainFormHandleEvent);
			break;

		default:
//			ErrFatalDisplay("Invalid Form Load Event");
			break;
		}
		return 1;
	}

	return false;
}

int TimeUntilNextPeriod (void)
{
	int TimeRemaining;

	TimeRemaining = Game.NextPeriodTime - TimGetTicks();
	if (TimeRemaining < 0)
		TimeRemaining = 0;

	return TimeRemaining;
}

void AppEventLoop(void)
{
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent(&event, TimeUntilNextPeriod());

		// If it's time, advance the game state.  Ensures a standard, constant framerate on all hardware.
		if (TimeUntilNextPeriod() == 0)
			GameStateAdvance();

		// intercept the hard keys to prevent them from switching apps or clicking or otherwise being annoying.
		if ( event.eType == keyDownEvent && !(event.data.keyDown.modifiers & poweredOnKeyMask) )
		{
			if (  	event.data.keyDown.chr == pageUpChr
				||	event.data.keyDown.chr == pageDownChr )
				continue;
		}

		// intercept all pen movement to prevent Grafitti from taking over.
		if ( 	event.eType == penDownEvent 
			||	event.eType == penUpEvent 
			||	event.eType == penMoveEvent )
			continue;

		if (!SysHandleEvent(&event))
			if (!MenuHandleEvent(0, &event, &error))
				if (!AppHandleEvent(&event))
					FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}

Err AppStart(void)
{
	PongPreferenceType prefs;
	UInt16 prefsSize;
	UInt16 error;
	RectangleType bounds;

	// Read the saved preferences / saved-state information.
	prefsSize = sizeof(PongPreferenceType);
	PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, 1);

	Game.OffscreenWindow = WinCreateOffscreenWindow (ScreenWidth, ScreenHeight, screenFormat, &error);

	// Setup the screenBuffer.  Clear it for use.
	WinSetDrawWindow(Game.OffscreenWindow);
	bounds.topLeft.x = 0;
	bounds.topLeft.y = 0;
	bounds.extent.x = ScreenWidth;
	bounds.extent.y = ScreenHeight;
	WinEraseRectangle(&bounds, 0);

	FrmAlert( InstructionsAlert );

	NewGame();

	return 0;
}

void AppStop(void)
{
	PongPreferenceType prefs;

	// Write the saved preferences / saved-state information.  This data 
	// will be backed up during a HotSync.
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum, &prefs, sizeof (prefs), 1);

	FrmCloseAllForms();
}

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error;

	switch (cmd)
	{
	case sysAppLaunchCmdNormalLaunch:
		error = AppStart();
		if (error)
			return error;

		FrmGotoForm(MainForm);
		AppEventLoop();

		AppStop();
		break;

	default:
		break;
	}

	return errNone;
}

